from memsql.common.connection_pool import MySQLError
from memsql.common import errorcodes, sql_utility
import time
import uuid

LOCK_TABLE = """\
CREATE TABLE IF NOT EXISTS %(name)s (
    id VARCHAR(255) PRIMARY KEY,
    lock_hash BINARY(32),
    owner VARCHAR(1024),
    last_contact TIMESTAMP DEFAULT NOW(),
    expiry INT
)"""

class SQLLockManager(sql_utility.SQLUtility):
    def __init__(self, table_prefix="sqllock"):
        """ Initialize the SQLLockManager with the specified table prefix.
        """
        super(SQLLockManager, self).__init__()

        self.table_name = table_prefix.rstrip('_') + '_locks'
        self._define_table(self.table_name, LOCK_TABLE % { 'name': self.table_name })

    ###############################
    # Public Interface

    def acquire(self, lock_id, owner='', expiry=5 * 60, block=False, timeout=None, retry_interval=0.5):
        start = time.time()
        while 1:
            lock_ref = self._acquire_lock(lock_id, owner, expiry)
            if lock_ref is None and block:
                if timeout is not None and (time.time() - start) > timeout:
                    break
                time.sleep(retry_interval)
            else:
                break
        return lock_ref

    ###############################
    # Private Interface

    def _acquire_lock(self, lock_id, owner, expiry):
        try:
            with self._db_conn() as conn:
                conn.execute('''
                    DELETE FROM %(table_name)s
                    WHERE last_contact <= NOW() - expiry
                ''' % { 'table_name': self.table_name })

                lock_hash = uuid.uuid1().hex
                conn.execute('''
                    INSERT INTO %s (id, lock_hash, owner, expiry)
                    VALUES (%%s, %%s, %%s, %%s)
                ''' % self.table_name, lock_id, lock_hash, owner, expiry)

                return SQLLock(lock_id=lock_id, lock_hash=lock_hash, owner=owner, manager=self)

        except MySQLError as (errno, msg):
            if errno == errorcodes.ER_DUP_ENTRY:
                return None
            else:
                raise

class SQLLock(object):
    def __init__(self, lock_id, lock_hash, owner, manager):
        self._manager = manager
        self._lock_id = lock_id
        self._lock_hash = lock_hash

        self.owner = owner

    ###############################
    # Public Interface

    def valid(self):
        with self._db_conn() as conn:
            row = conn.get('''
                SELECT
                    (lock_hash=%%s && last_contact > NOW() - expiry) AS valid
                FROM %s WHERE id = %%s
            ''' % self._manager.table_name, self._lock_hash, self._lock_id)

        return bool(row is not None and row.valid)

    def ping(self):
        """ Notify the manager that this lock is still active. """
        with self._db_conn() as conn:
            affected_rows = conn.query('''
                UPDATE %s
                SET last_contact=NOW()
                WHERE id = %%s AND lock_hash = %%s
            ''' % self._manager.table_name, self._lock_id, self._lock_hash)

        return bool(affected_rows == 1)

    def release(self):
        """ Release the lock. """
        if self.valid():
            with self._db_conn() as conn:
                affected_rows = conn.query('''
                    DELETE FROM %s
                    WHERE id = %%s AND lock_hash = %%s
                ''' % self._manager.table_name, self._lock_id, self._lock_hash)
            return bool(affected_rows == 1)
        else:
            return False

    ###############################
    # Context Management

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.release()

    ###############################
    # Public Interface

    def _db_conn(self):
        return self._manager._db_conn()



from memsql.common import json, sql_utility, database, util
import sys
import time
import random
import uuid
import copy
import types
import calendar
from datetime import datetime
from dateutil import parser
from contextlib import contextmanager

PRIMARY_TABLE = """\
CREATE TABLE IF NOT EXISTS %(table_name)s (
    id BIGINT AUTO_INCREMENT PRIMARY KEY,
    created INT NOT NULL,

    data JSON,

    execution_id CHAR(32) DEFAULT NULL,
    steps JSON,

    started INT DEFAULT 0 NOT NULL,
    last_contact INT DEFAULT 0 NOT NULL,
    update_count INT DEFAULT 0 NOT NULL,
    finished INT DEFAULT 0 NOT NULL,

    result AS data::$result PERSISTED VARCHAR(255),

    INDEX (created),
    INDEX (started),
    INDEX (last_contact)
)"""

class SQLStepQueueException(Exception):
    pass

class TaskDoesNotExist(SQLStepQueueException):
    pass

class StepAlreadyStarted(SQLStepQueueException):
    pass

class StepNotStarted(SQLStepQueueException):
    pass

class StepAlreadyFinished(SQLStepQueueException):
    pass

class AlreadyFinished(SQLStepQueueException):
    pass

class StepRunning(SQLStepQueueException):
    pass

class SQLStepQueue(sql_utility.SQLUtility):
    def __init__(self, table_name, execution_ttl=60):
        """
        table_name      the table name for the queue in the DB
        execution_ttl   the amount of time (in seconds) that can pass before a task is automatically requeued
        """
        super(SQLStepQueue, self).__init__()

        self.table_name = table_name
        self.execution_ttl = execution_ttl
        self._define_table(self.table_name, PRIMARY_TABLE % { 'table_name': self.table_name })

    ###############################
    # Public Interface

    def qsize(self, extra_predicate=None):
        """ Return an approximate number of queued tasks in the queue. """
        count = self._query_queued('COUNT(*) AS count', extra_predicate=extra_predicate)
        return count[0].count

    def enqueue(self, data):
        """ Enqueue task with specified data. """
        jsonified_data = json.dumps(data)
        with self._db_conn() as conn:
            return conn.execute(
                'INSERT INTO %s (created, data) VALUES (UNIX_TIMESTAMP(), %%(data)s)' % self.table_name,
                data=jsonified_data
            )

    def start(self, block=False, timeout=None, retry_interval=0.5, extra_predicate=None):
        """
        Retrieve a task handler from the queue.

        If block is True, this function will block until it is able to retrieve a task.
        If block is True and timeout is a number it will block for at most <timeout> seconds.
        retry_interval is the maximum time in seconds between successive retries.

        extra_predicate
        If extra_predicate is defined, it should be a tuple of (raw_predicate, predicate_args)
        raw_predicate will be prefixed by AND, and inserted into the WHERE condition in the queries.
        predicate_args will be sql escaped and formatted into raw_predicate.
        """
        start = time.time()
        while 1:
            task_handler = self._dequeue_task(extra_predicate)
            if task_handler is None and block:
                if timeout is not None and (time.time() - start) > timeout:
                    break
                time.sleep(retry_interval * (random.random() + 0.1))
            else:
                break
        return task_handler

    def bulk_finish(self, result='cancelled', extra_predicate=None):
        extra_predicate_sql = self._build_extra_predicate(extra_predicate)

        with self._db_conn() as conn:
            num_affected = conn.query('''
                UPDATE %s
                SET
                    execution_id = 0,
                    last_contact = UNIX_TIMESTAMP(),
                    update_count = update_count + 1,
                    steps = '[]',
                    started = UNIX_TIMESTAMP(),
                    finished = UNIX_TIMESTAMP(),
                    data::$result = %%(result)s
                WHERE
                    finished = 0
                    AND (
                        execution_id IS NULL
                        OR last_contact <= UNIX_TIMESTAMP() - %%(ttl)s
                    )
                    %s
            ''' % (self.table_name, extra_predicate_sql),
                result=result,
                ttl=self.execution_ttl)

        return num_affected

    def checkout(self, task_id, execution_id):
        return TaskHandler(execution_id=execution_id, task_id=task_id, queue=self)

    ###############################
    # Private Interface

    def _query_queued(self, projection, limit=None, extra_predicate=None):
        extra_predicate_sql = self._build_extra_predicate(extra_predicate)

        with self._db_conn() as conn:
            result = conn.query('''
                SELECT
                    %s
                FROM %s
                WHERE
                    finished = 0
                    AND (
                        execution_id IS NULL
                        OR last_contact <= UNIX_TIMESTAMP() - %%(ttl)s
                    )
                    %s
                ORDER BY created ASC
                LIMIT %%(limit)s
            ''' % (projection, self.table_name, extra_predicate_sql),
                ttl=self.execution_ttl,
                limit=sys.maxint if limit is None else limit)
        return result

    def _dequeue_task(self, extra_predicate=None):
        execution_id = uuid.uuid1().hex

        extra_predicate_sql = self._build_extra_predicate(extra_predicate)

        with self._db_conn() as conn:
            while True:
                possible_tasks = self._query_queued('id, created, data', limit=5, extra_predicate=extra_predicate)

                if not possible_tasks:
                    # nothing to dequeue
                    return None

                for possible_task in possible_tasks:
                    # attempt to claim the task
                    success = conn.query('''
                        UPDATE %s
                        SET
                            execution_id = %%(execution_id)s,
                            last_contact = UNIX_TIMESTAMP(),
                            update_count = update_count + 1,
                            started = UNIX_TIMESTAMP(),
                            steps = '[]'
                        WHERE
                            id = %%(task_id)s
                            AND finished = 0
                            AND (
                                execution_id IS NULL
                                OR last_contact <= UNIX_TIMESTAMP() - %%(ttl)s
                            )
                            %s
                    ''' % (self.table_name, extra_predicate_sql),
                        execution_id=execution_id,
                        task_id=possible_task.id,
                        ttl=self.execution_ttl)

                    if success == 1:
                        return TaskHandler(execution_id=execution_id, task_id=possible_task.id, queue=self)

    def _build_extra_predicate(self, extra_predicate):
        """ This method is a good one to extend if you want to create a queue which always applies an extra predicate. """
        if extra_predicate is None:
            return ''

        # if they don't have a supported format seq, wrap it for them
        if not isinstance(extra_predicate[1], (types.ListType, types.DictType, types.TupleType)):
            extra_predicate = [extra_predicate[0], (extra_predicate[1], )]

        extra_predicate = database.escape_query(*extra_predicate)

        return 'AND (' + extra_predicate + ')'

class TaskHandler(object):
    def __init__(self, execution_id, task_id, queue):
        self.execution_id = execution_id
        self.task_id = task_id
        self._queue = queue

        self.started = 0
        self.finished = 0
        self.data = None
        self.steps = None

        self._refresh()

    ###############################
    # Public Interface

    def valid(self):
        """ Check to see if we are still active. """
        if self.finished != 0:
            return False

        with self._db_conn() as conn:
            row = conn.get('''
                SELECT (last_contact > UNIX_TIMESTAMP() - %%(ttl)s) AS valid
                FROM %s
                WHERE
                    id = %%(task_id)s
                    AND execution_id = %%(execution_id)s
            ''' % self._queue.table_name,
                ttl=self._queue.execution_ttl,
                task_id=self.task_id,
                execution_id=self.execution_id)

        return bool(row is not None and row.valid)

    def ping(self):
        """ Notify the queue that this task is still active. """
        if self.finished != 0:
            raise AlreadyFinished()

        with self._db_conn() as conn:
            success = conn.query('''
                UPDATE %s
                SET
                    last_contact=UNIX_TIMESTAMP(),
                    update_count=update_count + 1
                WHERE
                    id = %%(task_id)s
                    AND execution_id = %%(execution_id)s
                    AND last_contact > UNIX_TIMESTAMP() - %%(ttl)s
            ''' % self._queue.table_name,
                task_id=self.task_id,
                execution_id=self.execution_id,
                ttl=self._queue.execution_ttl)

        if success != 1:
            raise TaskDoesNotExist()

    def finish(self, result='success'):
        if self._running_steps() != 0:
            raise StepRunning()
        if self.finished != 0:
            raise AlreadyFinished()

        data = copy.deepcopy(self.data)
        data['result'] = result
        self._save(data=data, finished=calendar.timegm(datetime.utcnow().timetuple()))

    def requeue(self):
        if self._running_steps() != 0:
            raise StepRunning()
        if self.finished != 0:
            raise AlreadyFinished()

        with self._db_conn() as conn:
            success = conn.query('''
                UPDATE %s
                SET
                    last_contact=0,
                    update_count=update_count + 1,
                    started=0,
                    steps=NULL,
                    execution_id=NULL,
                    finished=0,
                    data=JSON_DELETE_KEY(data, 'result')
                WHERE
                    id = %%(task_id)s
                    AND execution_id = %%(execution_id)s
                    AND last_contact > UNIX_TIMESTAMP() - %%(ttl)s
            ''' % self._queue.table_name,
                task_id=self.task_id,
                execution_id=self.execution_id,
                ttl=self._queue.execution_ttl)

        if success != 1:
            raise TaskDoesNotExist()

    def start_step(self, step_name):
        """ Start a step. """
        if self.finished != 0:
            raise AlreadyFinished()

        step_data = self._get_step(step_name)
        if step_data is not None:
            if 'stop' in step_data:
                raise StepAlreadyFinished()
            else:
                raise StepAlreadyStarted()

        steps = copy.deepcopy(self.steps)
        steps.append({
            "start": datetime.utcnow(),
            "name": step_name
        })
        self._save(steps=steps)

    def stop_step(self, step_name):
        """ Stop a step. """
        if self.finished != 0:
            raise AlreadyFinished()

        steps = copy.deepcopy(self.steps)

        step_data = self._get_step(step_name, steps=steps)
        if step_data is None:
            raise StepNotStarted()
        elif 'stop' in step_data:
            raise StepAlreadyFinished()

        step_data['stop'] = datetime.utcnow()

        step_data['duration'] = util.timedelta_total_seconds(step_data['stop'] - step_data['start'])
        self._save(steps=steps)

    @contextmanager
    def step(self, step_name):
        self.start_step(step_name)
        yield
        self.stop_step(step_name)

    def refresh(self):
        self._refresh()

    def save(self):
        self._save()

    ###############################
    # Private Interface

    def _get_step(self, step_name, steps=None):
        for step in (steps if steps is not None else self.steps):
            if step['name'] == step_name:
                return step
        return None

    def _running_steps(self):
        return len([s for s in self.steps if 'stop' not in s])

    def _db_conn(self):
        return self._queue._db_conn()

    def _refresh(self):
        with self._db_conn() as conn:
            row = conn.get('''
                SELECT * FROM %s
                WHERE
                    id = %%(task_id)s
                    AND execution_id = %%(execution_id)s
                    AND last_contact > UNIX_TIMESTAMP() - %%(ttl)s
            ''' % self._queue.table_name,
                task_id=self.task_id,
                execution_id=self.execution_id,
                ttl=self._queue.execution_ttl)

        if not row:
            raise TaskDoesNotExist()

        self.task_id = row.id
        self.data = json.loads(row.data)
        self.steps = self._load_steps(json.loads(row.steps))
        self.started = row.started
        self.finished = row.finished

    def _load_steps(self, raw_steps):
        """ load steps -> basically load all the datetime isoformats into datetimes """
        for step in raw_steps:
            if 'start' in step:
                step['start'] = parser.parse(step['start'])
            if 'stop' in step:
                step['stop'] = parser.parse(step['stop'])
        return raw_steps

    def _save(self, finished=None, steps=None, data=None):
        with self._db_conn() as conn:
            success = conn.query('''
                UPDATE %s
                SET
                    last_contact=UNIX_TIMESTAMP(),
                    update_count=update_count + 1,
                    steps=%%(steps)s,
                    finished=%%(finished)s,
                    data=%%(data)s
                WHERE
                    id = %%(task_id)s
                    AND execution_id = %%(execution_id)s
                    AND last_contact > UNIX_TIMESTAMP() - %%(ttl)s
            ''' % self._queue.table_name,
                task_id=self.task_id,
                execution_id=self.execution_id,
                ttl=self._queue.execution_ttl,
                steps=json.dumps(steps if steps is not None else self.steps),
                finished=finished if finished is not None else self.finished,
                data=json.dumps(data if data is not None else self.data))

        if success != 1:
            raise TaskDoesNotExist()
        else:
            if steps is not None:
                self.steps = steps
            if finished is not None:
                self.finished = finished
            if data is not None:
                self.data = data



from memsql.common import sql_utility, database, errorcodes
from wraptor.context import timer
import math

DATA_TABLE = """\
CREATE /*!90618 REFERENCE */ TABLE IF NOT EXISTS %(name)s (
    name VARCHAR(255) PRIMARY KEY,
    data LONGTEXT
)"""

class NetworkTester(sql_utility.SQLUtility):
    def __init__(self, table_prefix='network_tester', payload_size=1024 * 100):
        super(NetworkTester, self).__init__()

        self._payload_size = payload_size
        self.table_prefix = table_prefix.rstrip('_')
        self.data_table = self.table_prefix + '_data'
        self._define_table(self.data_table, DATA_TABLE % { 'name': self.data_table })

    ###############################
    # Public Interface

    def setup(self):
        super(NetworkTester, self).setup()

        self.half_payload = 'a' * (self._payload_size / 2)
        self.payload = self.half_payload * 2

        # bootstrap table
        with self._db_conn() as conn:
            try:
                conn.execute('''
                    INSERT INTO %s VALUES
                    ("half_payload", %%s),
                    ("payload", %%s)
                ''' % self.data_table, self.half_payload, self.payload)
            except database.MySQLError as (errno, msg):
                if errno == errorcodes.ER_DUP_KEY:
                    pass
                else:
                    raise

    def estimate_latency(self, iterations=100):
        results = []
        with self._db_conn() as conn:
            for _ in range(iterations):
                with timer() as t:
                    conn.ping()
                results.append(t.interval)

        computed = self._compute_results(results)
        return computed

    def estimate_roundtrip(self, iterations=100):
        results = []
        query = 'SELECT data AS a FROM %s # %%s' % (self.data_table)
        with self._db_conn() as conn:
            half_payload = conn.get('SELECT data FROM %s WHERE name="half_payload"' % self.data_table).data
            conn.execute(query, half_payload)  # ensure compile

            for _ in range(iterations):
                with timer() as t:
                    conn.query(query, half_payload)
                results.append((len(half_payload) * 2) / (t.interval / 1000.))

        computed = self._compute_results(results)
        return computed

    def estimate_upload(self, iterations=100):
        results = []
        query = '!!%s'
        with self._db_conn() as conn:
            payload = conn.get('SELECT data FROM %s WHERE name="payload"' % self.data_table).data

            for _ in range(iterations):
                try:
                    with timer() as t:
                        conn.query(query, payload)
                except database.MySQLError as (errno, msg):
                    if errno == errorcodes.ER_PARSE_ERROR:
                        pass
                    else:
                        raise
                results.append(len(payload) / (t.interval / 1000.))

        computed = self._compute_results(results)
        return computed

    def estimate_download(self, iterations=100):
        results = []
        query = 'SELECT data AS a FROM %s WHERE name="payload"' % (self.data_table)
        with self._db_conn() as conn:
            for _ in range(iterations):
                with timer() as t:
                    payload = conn.get(query).a
                results.append(len(payload) / (t.interval / 1000.))

        computed = self._compute_results(results)
        return computed

    ###############################
    # Private Interface

    def _compute_results(self, data):
        data.sort()

        return {
            '99th': self._percentile(data, 0.997),
            '95th': self._percentile(data, 0.95),
            '68th': self._percentile(data, 0.68),
            'median': self._percentile(data, 0.50),
            'avg': sum(data) / len(data),
            'max': max(data),
            'min': min(data),
        }

    def _percentile(self, data, percent):
        """
        Find the percentile of a sorted list of values.
        From: http://stackoverflow.com/questions/2374640/how-do-i-calculate-percentiles-with-python-numpy

        data: sorted list of values
        percent: the percentile to compute. (range: 0, 1)
        """
        if not data:
            return None
        k = (len(data) - 1) * percent
        f = math.floor(k)
        c = math.ceil(k)
        if f == c:
            return data[int(k)]
        d0 = data[int(f)] * (c - k)
        d1 = data[int(c)] * (k - f)
        return d0 + d1



from memsql.common.connection_pool import ConnectionPool, PoolConnectionException
from memsql.common import errorcodes
from memsql.common.database import DatabaseError
from wraptor.decorators import memoize
import threading
import random
import logging

class RandomAggregatorPool(object):
    """ A automatic fail-over connection pool.

    One layer above the connection pool. It's purpose is to choose a
    random aggregator and use it while it is available. If not it fails
    over to another aggregator.  The class maintains the list of
    aggregators by periodically calling `SHOW AGGREGATORS`.

    Note: If you point this class at a MemSQL Singlebox instance, it
    will still work, but all connections will just be made to the
    singlebox node.
    """

    def __init__(self, host, port, user='root', password='', database='information_schema'):
        """ Initialize the RandomAggregatorPool with connection
        information for an aggregator in a MemSQL Distributed System.

        All aggregator connections will share the same user/password/database.
        """
        self.logger = logging.getLogger('memsql.random_aggregator_pool')
        self._pool = ConnectionPool()
        self._refresh_aggregator_list = memoize(30)(self._update_aggregator_list)
        self._lock = threading.RLock()

        self._primary_aggregator = (host, port)
        self._user = user
        self._password = password
        self._database = database
        self._aggregators = []
        self._aggregator = None
        self._master_aggregator = None

    def connect(self):
        """ Returns an aggregator connection, and periodically updates the aggregator list. """
        conn = self._connect()
        self._refresh_aggregator_list(conn)
        return conn

    def connect_master(self):
        if self._master_aggregator is None:
            with self._pool_connect(self._primary_aggregator) as conn:
                self._update_aggregator_list(conn)
                conn.expire()
        try:
            return self._pool_connect(self._master_aggregator)
        except PoolConnectionException:
            return None

    def close(self):
        self._pool.close()

    def _pool_connect(self, agg):
        """ `agg` should be (host, port)
            Returns a live connection from the connection pool
        """
        return self._pool.connect(agg[0], agg[1], self._user, self._password, self._database)

    def _connect(self):
        """ Returns an aggregator connection. """
        with self._lock:
            if self._aggregator:
                try:
                    return self._pool_connect(self._aggregator)
                except PoolConnectionException:
                    self._aggregator = None

            if not len(self._aggregators):
                with self._pool_connect(self._primary_aggregator) as conn:
                    self._update_aggregator_list(conn)
                    conn.expire()

            random.shuffle(self._aggregators)

            last_exception = None
            for aggregator in self._aggregators:
                self.logger.debug('Attempting connection with %s:%s' % (aggregator[0], aggregator[1]))

                try:
                    conn = self._pool_connect(aggregator)
                    # connection successful!
                    self._aggregator = aggregator
                    return conn
                except PoolConnectionException as e:
                    # connection error
                    last_exception = e
            else:
                # bad news bears...  try again later
                self._aggregator = None
                self._aggregators = []

                raise last_exception

    def _update_aggregator_list(self, conn):
        try:
            rows = conn.query('SHOW AGGREGATORS')
        except DatabaseError as e:
            if e.args[0] == errorcodes.ER_DISTRIBUTED_NOT_AGGREGATOR:
                # connected to memsql singlebox
                self._aggregators = [self._primary_aggregator]
                self._master_aggregator = self._primary_aggregator
            else:
                raise
        else:
            with self._lock:
                self._aggregators = []
                for row in rows:
                    if row.Host == '127.0.0.1':
                        # this is the aggregator we are connecting to
                        row['Host'] = conn.connection_info()[0]
                    if int(row.Master_Aggregator) == 1:
                        self._master_aggregator = (row.Host, row.Port)
                    self._aggregators.append((row.Host, row.Port))

            assert len(self._aggregators) > 0, "Failed to retrieve a list of aggregators"

        self.logger.debug('Aggregator list is updated to %s. Current aggregator is %s.' % (self._aggregators, self._aggregator))



import simplejson

def simplejson_datetime_serializer(obj):
    """
    Designed to be passed as the default kwarg in simplejson.dumps.  Serializes dates and datetimes to ISO strings.
    """
    if hasattr(obj, 'isoformat'):
        return obj.isoformat()
    else:
        raise TypeError('Object of type %s with value of %s is not JSON serializable' % (type(obj), repr(obj)))

def _set_defaults(kwargs):
    kwargs.setdefault('separators', (',', ':'))
    kwargs.setdefault('default', simplejson_datetime_serializer)
    return kwargs

def dumps(data, **kwargs):
    return simplejson.dumps(data, **_set_defaults(kwargs))

def loads(data, **kwargs):
    return simplejson.loads(data, **kwargs)



from memsql.common.connection_pool import ConnectionPool
from memsql.common import exceptions

class SQLUtility(object):
    def __init__(self):
        self._pool = ConnectionPool()
        self._db_args = None
        self._tables = {}

    ###############################
    # Public Interface

    def connect(self, host='127.0.0.1', port=3306, user='root', password='', database=None):
        """ Connect to the database specified """

        if database is None:
            raise exceptions.RequiresDatabase()

        self._db_args = { 'host': host, 'port': port, 'user': user, 'password': password, 'database': database }
        with self._db_conn() as conn:
            conn.query('SELECT 1')
        return self

    def disconnect(self):
        self._pool.close()

    def setup(self):
        """ Initialize the required tables in the database """
        with self._db_conn() as conn:
            for table_defn in self._tables.values():
                conn.execute(table_defn)
        return self

    def destroy(self):
        """ Destroy the SQLStepQueue tables in the database """
        with self._db_conn() as conn:
            for table_name in self._tables:
                conn.execute('DROP TABLE IF EXISTS %s' % table_name)
        return self

    def ready(self):
        """ Returns True if the tables have been setup, False otherwise """
        with self._db_conn() as conn:
            tables = [row.t for row in conn.query('''
                SELECT table_name AS t FROM information_schema.tables
                WHERE table_schema=%s
            ''', self._db_args['database'])]
        return all([table_name in tables for table_name in self._tables])

    ###############################
    # Protected Interface

    def _define_table(self, table_name, table_definition):
        self._tables[table_name] = table_definition

    def _db_conn(self):
        if self._db_args is None:
            raise exceptions.NotConnected()
        return self._pool.connect(**self._db_args)



def simple_expression(joiner=', ', **fields):
    """ Build a simple expression ready to be added onto another query.

    >>> simple_expression(joiner=' AND ', name='bob', role='admin')
    "`name`=%(_QB_name)s AND `name`=%(_QB_role)s", { '_QB_name': 'bob', '_QB_role': 'admin' }
    """
    expression, params = [], {}

    for field_name, value in sorted(fields.iteritems(), key=lambda (n, v): n):
        key = '_QB_%s' % field_name
        expression.append('`%s`=%%(%s)s' % (field_name, key))
        params[key] = value

    return joiner.join(expression), params

def update(table_name, **fields):
    """ Build a update query.

    >>> update('foo_table', a=5, b=2)
    "UPDATE `foo_table` SET `a`=%(_QB_a)s, `b`=%(_QB_b)s", { '_QB_a': 5, '_QB_b': 2 }
    """
    prefix = "UPDATE `%s` SET " % table_name
    sets, params = simple_expression(', ', **fields)
    return prefix + sets, params

def multi_insert(table_name, *rows):
    """ Build a multi-insert query.
        Each row in rows should be a dict of { column_name: column_value }

    >>> multi_insert('foo_table', { 'a': 5, 'b': 2 }, { 'a': 5, 'b': 2 })
    "INSERT INTO `foo_table` (`a`, `b`) VALUES (%(_QB_ROW_0)s), (%(_QB_ROW_1)s)"
    """
    return __multi_insert(table_name, rows)

def multi_replace(table_name, *rows):
    """ Build a multi-replace query.
        Each row in rows should be a dict of { column_name: column_value }

    >>> multi_replace('foo_table', { 'a': 5, 'b': 2 }, { 'a': 5, 'b': 2 })
    "REPLACE INTO `foo_table` (`a`, `b`) VALUES (%(_QB_ROW_0)s), (%(_QB_ROW_1)s)"
    """
    return __multi_insert(table_name, rows, replace=True)

def __multi_insert(table_name, rows, replace=False):
    cols = sorted(rows[0].keys())
    prefix = '%s INTO `%s` (%s) VALUES ' % (
        'REPLACE' if replace else 'INSERT',
        table_name,
        ', '.join(['`%s`' % col for col in cols])
    )
    sql, params = [], {}

    for i, row in enumerate(rows):
        key = '_QB_ROW_%d' % i
        params[key] = [ v for c, v in sorted(row.iteritems(), key=lambda (c, v): cols.index(c)) ]
        sql.append('(%%(%s)s)' % key)

    return prefix + ', '.join(sql), params




import _mysql, errno, Queue
import multiprocessing
import logging
from memsql.common import database

MySQLError = database.MySQLError
QUEUE_SIZE = 128

class PoolConnectionException(IOError):
    """ This exception consolidates all connection exceptions into one thing """

    def __init__(self, errno, message, connection_key):
        IOError.__init__(self, errno, message)
        (self.host, self.port, self.user, self.password, self.db_name, self.pid) = connection_key

    def _get_message(self):
        return self.args[1]

    message = property(_get_message)

class ConnectionPool(object):
    def __init__(self):
        self.logger = logging.getLogger('memsql.connection_pool')
        self._connections = {}
        self._fairies = {}

    def connect(self, host, port, user, password, database):
        current_proc = multiprocessing.current_process()
        key = (host, port, user, password, database, current_proc.pid)

        if key not in self._connections:
            self._connections[key] = Queue.Queue(maxsize=QUEUE_SIZE)

        fairy = _PoolConnectionFairy(key, self)
        fairy.connect()
        self._fairies[fairy] = 1
        return fairy

    def checkin(self, fairy, key, conn, expire_connection=False):
        if key not in self._connections:
            self._connections[key] = Queue.Queue(maxsize=QUEUE_SIZE)

        if expire_connection:
            try:
                conn.close()
            except Exception:
                self.logger.error("Could not close connection after fairy expired")
        else:
            try:
                self._connections[key].put_nowait(conn)
            except Queue.Full:
                conn.close()

        if fairy in self._fairies:
            del(self._fairies[fairy])

    def close(self):
        for fairy in self._fairies.keys():
            fairy.close()
        for queue in self._connections.values():
            while True:
                try:
                    conn = queue.get_nowait()
                    conn.close()
                except Queue.Empty:
                    break

    def size(self):
        """ Returns the number of connections cached by the pool. """
        return sum([q.qsize() for q in self._connections.values()]) + len(self._fairies)

class _PoolConnectionFairy(object):
    def __init__(self, key, pool):
        self._key = key
        self._pool = pool
        self._expired = False

    def expire(self):
        self._expired = True

    def close(self):
        self._pool.checkin(self, self._key, self._conn, expire_connection=self._expired)

    def connection_info(self):
        return (self._key[0], self._key[1])

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __wrap_errors(self, fn, *args, **kwargs):
        def wrapped(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except IOError as e:
                if e.errno in [errno.ECONNRESET, errno.ECONNREFUSED, errno.ETIMEDOUT]:
                    # socket connection issues
                    self.__handle_connection_failure(e)
                else:
                    raise
            except _mysql.OperationalError as e:
                # _mysql specific database connect issues, internal state issues
                if hasattr(self, '_conn'):
                    self.__potential_connection_failure(e)
                else:
                    self.__handle_connection_failure(e)
        return wrapped

    def __potential_connection_failure(self, e):
        """ OperationalError's are emitted by the _mysql library for
        almost every error code emitted by MySQL.  Because of this we
        verify that the error is actually a connection error before
        terminating the connection and firing off a PoolConnectionException
        """
        try:
            self._conn.query('SELECT 1')
        except (IOError, _mysql.OperationalError):
            # ok, it's actually an issue.
            self.__handle_connection_failure(e)
        else:
            # seems ok, probably programmer error
            raise _mysql.DatabaseError(*e.args)

    def __handle_connection_failure(self, e):
        # expire the connection so we don't return it to the pool accidentally
        self.expire()

        # build and raise the new consolidated exception
        message = e.message
        if isinstance(e, _mysql.OperationalError) or (hasattr(e, 'args') and len(e.args) >= 2):
            err_num = e.args[0]
            message = e.args[1]
        elif hasattr(e, 'errno'):
            err_num = e.errno
        else:
            err_num = errno.ECONNABORTED

        raise PoolConnectionException(err_num, message, self._key)

    ##################
    ## Wrap DB Api to deal with connection issues and so on in an intelligent way

    def connect(self):
        try:
            self._conn = self._pool._connections[self._key].get_nowait()
        except Queue.Empty:
            (host, port, user, password, db_name, pid) = self._key
            _connect = self.__wrap_errors(database.connect)
            self._conn = _connect(host=host, port=port, user=user, password=password, database=db_name)

    # catchall
    def __getattr__(self, key):
        method = getattr(self._conn, key, None)
        if method is None:
            raise AttributeError('Attribute `%s` does not exist' % key)
        else:
            return self.__wrap_errors(method)



#!/usr/bin/env python

from setuptools import setup
from setuptools.command.test import test as TestCommand

# get version
from memsql import __version__

class PyTest(TestCommand):
    user_options = [
        ('watch', 'w',
         "watch tests for changes"),
        ('pdb', 'i',
         "start pdb on failures"),
        ('scan=', 's',
         "only search for tests in the specified directory or file"),
        ('exitfirst', 'x',
         "Stop tests on first failure"),
        ('expression=', 'k',
         "Only run tests matching given expression"),
    ]
    boolean_options = ['watch']

    def initialize_options(self):
        self.watch = False
        self.pdb = False
        self.scan = None
        self.exitfirst = False
        self.expression = None
        self.test_suite = None
        self.test_module = None
        self.test_loader = None

    def finalize_options(self):
        TestCommand.finalize_options(self)

        self.test_suite = True
        self.test_args = ['-v']
        if self.watch:
            self.test_args.append('-f')
        if self.pdb:
            self.test_args.append('--pdb')
        if self.exitfirst:
            self.test_args.append('-x')
        if self.expression:
            self.test_args.extend(['-k', self.expression])
        if self.scan is not None:
            self.test_args.append(self.scan)
        else:
            self.test_args.append('memsql')

    def run_tests(self):
        import os, sys, glob

        MY_PATH = os.path.dirname(__file__)
        sys.path.append(MY_PATH)
        os.environ['PYTHONPATH'] = os.environ.get('PYTHONPATH', '') + ':' + MY_PATH

        egg_dirs = glob.glob('*.egg')
        ignore_args = ['--ignore=%s' % d for d in egg_dirs]

        import pytest
        errno = pytest.main(ignore_args + self.test_args)
        raise sys.exit(errno)

setup(
    name='memsql',
    version=__version__,
    author='MemSQL',
    author_email='support@memsql.com',
    url='http://github.com/memsql/memsql-python',
    license='LICENSE.txt',
    description='Useful utilities and plugins for MemSQL integration.',
    long_description=open('README.rst').read(),
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],
    packages=[
        'memsql',
        'memsql.common',
        'memsql.perf',
    ],
    zip_safe=False,
    install_requires=['MySQL-python==1.2.5', 'wraptor', 'simplejson', 'python-dateutil==2.2'],
    tests_require=['ordereddict', 'pytest', 'mock', 'pytest-xdist'],
    cmdclass={ 'test': PyTest },
)



"""A lightweight wrapper around _mysql."""

import _mysql
import itertools
import time
import operator

try:
    from thread import get_ident as _get_ident
except ImportError:
    from dummy_thread import get_ident as _get_ident

from MySQLdb.converters import conversions

MySQLError = _mysql.MySQLError
OperationalError = _mysql.OperationalError
DatabaseError = _mysql.DatabaseError

encoders = dict([ (k, v) for k, v in conversions.items() if not isinstance(k, int) ])

def connect(*args, **kwargs):
    return Connection(*args, **kwargs)

class Connection(object):
    """A lightweight wrapper around _mysql DB-API connections.

    The main value we provide is wrapping rows in a dict/object so that
    columns can be accessed by name. Typical usage::

        db = database.Connection("localhost", "mydatabase")
        for article in db.query("SELECT * FROM articles"):
            print article.title

    Cursors are hidden by the implementation, but other than that, the methods
    are very similar to the DB-API.

    We explicitly set the timezone to UTC and the character encoding to
    UTF-8 on all connections to avoid time zone and encoding errors.
    """
    def __init__(self, host, port=3306, database="information_schema", user=None, password=None,
                 max_idle_time=7 * 3600):
        self.max_idle_time = max_idle_time

        args = {
            "db": database,
            "conv": conversions
        }

        if user is not None:
            args["user"] = user
        if password is not None:
            args["passwd"] = password

        args["host"] = host
        args["port"] = int(port)

        self._db = None
        self._db_args = args

        self._last_use_time = time.time()
        self.reconnect()
        self._db.set_character_set("utf8")

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        """Closes this database connection."""
        if getattr(self, "_db", None) is not None:
            self._db.close()
            self._db = None

    def connected(self):
        if self._db is not None:
            try:
                self.ping()
                return True
            except _mysql.InterfaceError:
                return False
        return False

    def reconnect(self):
        """Closes the existing database connection and re-opens it."""
        conn = _mysql.connect(**self._db_args)
        if conn is not None:
            self.close()
            self._db = conn

    def select_db(self, database):
        self._db.select_db(database)
        self._db_args['db'] = database

    def ping(self):
        """ Ping the server """
        return self._db.ping()

    def debug_query(self, query, *parameters, **kwparameters):
        return self._query(query, parameters, kwparameters, debug=True)

    def query(self, query, *parameters, **kwparameters):
        """
        Query the connection and return the rows (or affected rows if not a
        select query).  Mysql errors will be propogated as exceptions.
        """
        return self._query(query, parameters, kwparameters)

    def get(self, query, *parameters, **kwparameters):
        """Returns the first row returned for the given query."""
        rows = self._query(query, parameters, kwparameters)
        if not rows:
            return None
        elif not isinstance(rows, list):
            raise MySQLError("Query is not a select query")
        elif len(rows) > 1:
            raise MySQLError("Multiple rows returned for Database.get() query")
        else:
            return rows[0]

    # rowcount is a more reasonable default return value than lastrowid,
    # but for historical compatibility execute() must return lastrowid.
    def execute(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the lastrowid from the query."""
        return self.execute_lastrowid(query, *parameters, **kwparameters)

    def execute_lastrowid(self, query, *parameters, **kwparameters):
        """Executes the given query, returning the lastrowid from the query."""
        self._execute(query, parameters, kwparameters)
        self._result = self._db.store_result()
        return self._db.insert_id()

    def _query(self, query, parameters, kwparameters, debug=False):
        self._execute(query, parameters, kwparameters, debug)
        self._result = self._db.use_result()
        if self._result is None:
            return self._rowcount
        fields = zip(*self._result.describe())[0]
        rows = list(self._result.fetch_row(0))
        return SelectResult(fields, rows)

    def _execute(self, query, parameters, kwparameters, debug=False):
        if parameters and kwparameters:
            raise ValueError('database.py querying functions can receive *args or **kwargs, but not both')

        query = escape_query(query, parameters or kwparameters)
        if debug:
            print query

        self._ensure_connected()
        self._db.query(query)
        self._rowcount = self._db.affected_rows()

    def _ensure_connected(self):
        # Mysql by default closes client connections that are idle for
        # 8 hours, but the client library does not report this fact until
        # you try to perform a query and it fails.  Protect against this
        # case by preemptively closing and reopening the connection
        # if it has been idle for too long (7 hours by default).
        if (self._db is None or (time.time() - self._last_use_time > self.max_idle_time)):
            self.reconnect()
        self._last_use_time = time.time()


class Row(object):
    """A fast, ordered, partially-immutable dictlike object (or objectlike dict)."""

    def __init__(self, fields, values):
        self._fields = fields
        self._values = list(values)

    def __getattr__(self, name):
        try:
            return self._values[self._fields.index(name)]
        except ValueError, IndexError:
            raise AttributeError(name)

    def __getitem__(self, name):
        try:
            return self._values[self._fields.index(name)]
        except ValueError, IndexError:
            raise KeyError(name)

    def __setitem__(self, name, value):
        try:
            self._values[self._fields.index(name)] = value
        except ValueError, IndexError:
            self._fields += (name,)
            self._values += (value,)

    def __contains__(self, name):
        return name in self._fields

    has_key = __contains__

    def __sizeof__(self, name):
        return len(self._fields)

    def __iter__(self):
        return self._fields.__iter__()

    def __len__(self):
        return self._fields.__len__()

    def keys(self):
        return list(self._fields)

    def get(self, name, default=None):
        try:
            return self.__getitem__(name)
        except KeyError:
            return default

    def values(self):
        return list(self._values)

    def items(self):
        return zip(self._fields, self._values)

    def iterkeys(self):
        return iter(self._fields)

    def itervalues(self):
        return iter(self._values)

    def iteritems(self):
        for i, key in enumerate(self._fields):
            yield (key, self._values[i])

    def __eq__(self, other):
        if isinstance(other, Row):
            return dict.__eq__(dict(self.iteritems()), other) and all(itertools.imap(operator.eq, self, other))
        return dict.__eq__(dict(self.iteritems()), other)

    def __ne__(self, other):
        return not self == other

    def __repr__(self, _repr_running={}):
        call_key = id(self), _get_ident()
        if call_key in _repr_running:
            return '...'
        _repr_running[call_key] = 1
        try:
            if not self:
                return '%s()' % (self.__class__.__name__,)
            return '%s(%r)' % (self.__class__.__name__, self.items())
        finally:
            del _repr_running[call_key]

    ## for simplejson.dumps()
    def _asdict(self):
        return dict(self)

    def nope(self, *args, **kwargs):
        raise NotImplementedError('This object is partially immutable. To modify it, call "foo = dict(foo)" first.')

    update = nope
    pop = nope
    setdefault = nope
    fromkeys = nope
    clear = nope
    __delitem__ = nope
    __reversed__ = nope

class SelectResult(list):
    def __init__(self, fieldnames, rows):
        self.fieldnames = tuple(fieldnames)
        self.rows = rows

        data = [Row(self.fieldnames, row) for row in self.rows]
        list.__init__(self, data)

    def width(self):
        return len(self.fieldnames)

    def __getitem__(self, i):
        if isinstance(i, slice):
            return SelectResult(self.fieldnames, self.rows[i])
        return list.__getitem__(self, i)

def escape_query(query, parameters):
    if parameters:
        if isinstance(parameters, (list, tuple)):
            query = query % tuple(map(_escape, parameters))
        elif isinstance(parameters, dict):
            query = query % dict(map(lambda (key, item): (key, _escape(item)), parameters.iteritems()))
        else:
            assert False, 'not sure what to do with parameters of type %s' % type(parameters)

    return _escape_unicode(query)

def _escape_unicode(param):
    if isinstance(param, unicode):
        return param.encode('utf-8')
    else:
        return param

def _escape(param):
    if isinstance(param, (list, tuple)):
        param = [_escape_unicode(p) for p in param]
        return ','.join(_mysql.escape_sequence(param, encoders))
    else:
        return _mysql.escape(_escape_unicode(param), encoders)
